import os
import libvirt

from virtManager.asyncjob import vmmAsyncJob
from gi.repository import Gtk

from virtinst import log
from virtinst import Importer

from .lib import uiutil
from .baseclass import vmmGObjectUI
from .storagebrowse import vmmStorageBrowser


class vmmImportDialog(vmmGObjectUI):
    _instance = None

    @classmethod
    def reset_instance(cls):
        cls._instance = None

    @classmethod
    def show_instance(cls, parentobj, conn):
        try:
            if cls._instance:
                cls._instance.close()
            cls.reset_instance()
            cls._instance = vmmImportDialog(conn)
            cls._instance.show(parentobj.topwin, conn)
        except Exception as e:  # pragma: no cover
            parentobj.err.show_err(_("Error launching import dialog: %s")
                                   % str(e))

    def __init__(self, conn):
        vmmGObjectUI.__init__(self, "importvm.ui", "vmm-import")
        self.conn = conn
        self._archive_path = None

        self.builder.connect_signals({
            "on_vmm_import_delete_event": self._delete_event,
            "on_import_cancel_clicked": self._cancel_clicked,
            "on_import_finish_clicked": self._finish_clicked,
            "on_browse_local_clicked": self._browse_local_clicked,
        })

        self._build_pool_combo(self.widget("pool-model"))
        folder = self.config.get_default_export_directory()
        if folder:
            self.widget("install-import-entry").set_text(folder)

    def _cleanup(self):
        self.conn = None
        self._archive_path = None


    ##############
    # Public API #
    ##############

    def show(self, parent, conn):
        log.debug("Showing importvm wizard")
        self._build_pool_combo(self.widget("pool-model"))
        self.conn = conn
        self.topwin.set_transient_for(parent)
        self.topwin.present()

    def close(self, ignore1=None, ignore2=None):
        log.debug("Closing importvm wizard")
        self.topwin.hide()
        return 1

    def _vm_removed_cb(self, _conn, vm):
        if self.vm == vm:
            self.close()


    #############
    # Listeners #
    #############

    def _delete_event(self, ignore1, ignore2):
        self.close()
        return 1

    def _cancel_clicked(self, src):
        ignore = src
        self.close()

    def _finish_clicked(self, src):
        ignore = src
        self._finish()
    
    def _browse_local_clicked(self, src):
        ingore = src
        self._browse_file()

    ###################
    # import handling #
    ###################

    def _validate(self, conn):
        vm_name = os.path.basename(self._archive_path).rsplit('.', 2)[0]
        try:
            vm = conn.lookupByName(vm_name)
            if vm:
                self.err.show_err(
                    _("Unable to import virtual machine"),
                    _("Please delete the following virtual machine if you want"
                      " to import new one: %s") % vm_name)
                return False

            path = self.widget("install-import-entry").get_text()
            if not os.path.exists(path):
                self.err.show_err(
                    _("Unable to import virtual machine"),
                    _("Please select correct path to file") % vm_name)
                return False

        except:
            return True

    def _async_import(self, importer, tar_path, pool_name, checksum):
        importer.start_importing(tar_path, pool_name, checksum)

        try:
            pool = self.conn.get_pool_by_name(pool_name)
            self.idle_add(pool.refresh)
        except Exception:  # pragma: no cover
            log.debug("Error looking up pool=%s for refresh after "
                    "VM import.", pool_name, exc_info=True)

    def _finish(self):
        pool_name = self.widget("pool-model-text").get_text()
        checksum = self.widget("checksum").get_active()
        conn = self.conn.get_backend()
        importer = Importer(conn)

        if not checksum:
            res = self.err.yes_no(_("Are you sure you want to import virtual "
                                    "machine without checking the checksum?"))
            if not res:
                return

        if checksum and not os.path.exists(f"{self._archive_path}.md5"):
            res = self.err.yes_no(_("MD5 file not found, continue without "
                                    "checking the checksum?"))
            if res:
                checksum = False
            if not res:
                return

        if self._validate(conn) == False:
            self.close()
            return

        vmmAsyncJob.simple_async(self._async_import,
                        [importer, self._archive_path, pool_name, checksum],
                        self, _("Importing virtual machine"),
                        (_("Uncompressing '%s' and downloading disks with "
                           "xml files "
                           "(this may take a while)") % self._archive_path),
                        (_("Error importing virtual "
                          "machine: '%s'") % self._archive_path))

        self.close()

    def _build_pool_combo(self, combo):
        uiutil.build_simple_combo(combo, 
                                  [[None, pool.get_xmlobj().name]
                                   for pool in self.conn.list_pools()])

    def _browse_file(self):
        start_folder = self.config.get_default_export_directory()
        file_path = self.err.browse_local(self.conn,
            dialog_name=_("Select virtual machine archive"),
            choose_button=Gtk.STOCK_OPEN,
            start_folder = start_folder,
            _type=("gz", _("GZIP files")))

        if file_path and os.path.splitext(file_path)[1] == ".gz":
            self.widget("install-import-entry").set_text(file_path)
            self._archive_path = file_path
