import datetime
import os
import shutil
import subprocess
from virtManager.asyncjob import vmmAsyncJob

from virtManager.lib import uiutil
from virtinst import log
from virtinst.devices.interface import DeviceInterface
from virtinst.guest import Guest
from virtinst.importer import _upload_disk
from virtinst.cloner import _get_shareable_msg
from virtinst.storage import StoragePool, StorageVolume
from .baseclass import vmmGObjectUI


class vmmRestoreBackupDialog(vmmGObjectUI):
    _instance = None

    @classmethod
    def reset_instance(cls):
        cls._instance = None

    @classmethod
    def show_instance(cls, parentobj, vm):
        try:
            if cls._instance:
                cls._instance.close()
            cls.reset_instance()
            cls._instance = vmmRestoreBackupDialog(vm)
            cls._instance.show(parentobj.topwin, vm.conn)
        except Exception as e:  # pragma: no cover
            parentobj.err.show_err(
                _("Error launching restore backup dialog: %s") % str(e))

    def __init__(self, vm):
        vmmGObjectUI.__init__(self, "restorebackup.ui", "vmm-restore-backup")
        self.vm = vm
        self._backup_folder = None
        self.output = None
        self.overwrite = False

        self.builder.connect_signals({
            "on_vmm_restore_delete_event": self._delete_event,
            "on_restore_cancel_clicked": self._cancel_clicked,
            "on_restore_finish_clicked": self._finish_clicked,
            "on_browse_local_clicked": self._browse_local_clicked,
        })
        self._populate_restore_combo(self.widget("restore-checkpoints"))
        self.widget("restore-name").set_text(self.vm.get_name())

    def _cleanup(self):
        self.vm = None
        self._backup_folder = None
        self.output = None


    ##############
    # Public API #
    ##############

    def show(self, parent, conn):
        log.debug("Showing restore wizard")
        self.conn = conn
        self.topwin.set_transient_for(parent)
        self.topwin.present()

    def close(self, ignore1=None, ignore2=None):
        log.debug("Closing restore wizard")
        self.topwin.hide()
        return 1

    def _vm_removed_cb(self, _conn, vm):
        if self.vm == vm:
            self.close()


    ###############
    # UI building #
    ###############

    def _populate_restore_combo(self, combo):
        self._backup_folder = self.vm.get_xmlobj(inactive=True).get_backup_folder()
        if self._backup_folder:
            checkpoints = os.path.join(self._backup_folder, "checkpoints")
            items = []
            counter = 1
            for checkpoint in os.listdir(checkpoints):
                path = os.path.join(checkpoints, checkpoint)
                m_time = os.path.getmtime(path)
                dt_m = datetime.datetime.fromtimestamp(m_time)
                view = f"{counter}: " + str(dt_m).rsplit(":", 1)[0]
                items.append([checkpoint, view])
                counter += 1

            items.sort(key=lambda x: int(x[1].split(":")[0]))
            uiutil.build_simple_combo(combo, items)


    #############
    # Listeners #
    #############

    def _delete_event(self, ignore1, ignore2):
        self.close()
        return 1

    def _cancel_clicked(self, src):
        ignore = src
        self.close()

    def _finish_clicked(self, src):
        ignore = src
        self._finish()

    def _browse_local_clicked(self, src):
        ingore = src
        self._browse_file()

    ####################
    # restore handling #
    ####################

    def _build_command(self):
        command = (f"virtnbdrestore -U {self.conn.get_uri()} -o {self.output} "
                   f"-i {self._backup_folder}")

        checkpoint = uiutil.get_list_selection(self.widget("restore-checkpoints"))
        if checkpoint:
            checkpoint = checkpoint.rsplit(".", 1)[0]
            command += f" --until {checkpoint}"
        return command

    def _async_restore(self, new_name):
        vm_name = self.vm.get_name()
        self.output = os.path.join("/tmp/restore", vm_name)
        command = self._build_command()
        vm = None
        disks = []
        try:
            pool = None
            for disk in self.vm.xmlobj.devices.disk:
                conn = self.conn.get_backend()
                vol = conn.storageVolLookupByPath(disk.get_source_path())
                if vol:
                    pool = vol.storagePoolLookupByVolume()
            if pool is None:
                raise

            subprocess.run(command, shell=True,
                                    capture_output=True, text=True)
            if self.overwrite:
                self.vm.delete()
                for disk in self.vm.xmlobj.devices.disk:
                    if not _get_shareable_msg(disk):
                        vol = conn.storageVolLookupByPath(disk.source.file)
                        vol.delete(0)

            files = os.listdir(self.output)
            config = list(filter(lambda file: file.endswith(".xml"), files))[0]
            self.save_xml(conn, new_name, config)
            for file in os.listdir(self.output):
                if file.endswith(".qcow2"):
                    vol_name = self.save_disk(conn, pool, file, new_name)
                    disks.append(vol_name)
            vm = conn.defineXML(self._new_guest.get_xml())

        except Exception as e:
            log.debug("Restoring failed: %s", str(e))
            if vm:
                vm.undefine()
            if disks:
                for disk in disks:
                    pool = StoragePool(conn, pool.XMLDesc(0))
                    path = os.path.join(pool.target_path, disk)
                    vol = conn.storageVolLookupByPath(path)
                    vol.delete(0)
        finally:
            if os.path.exists(self.output):
                shutil.rmtree(self.output)

        log.debug("Restore finished.")

    def _finish(self):
        new_name = self.widget("restore-name").get_text()
        conn = self.conn.get_backend()
        dom = None
        try:
            dom = conn.lookupByName(new_name)
        except:
            pass
        if dom and new_name != self.vm.get_name():
            res = self.err.show_err(
                _("Virtual machine with the same name already exists"))
            return
        if new_name == self.vm.get_name():
            res = self.err.yes_no(
                _("Are you sure you want to restore virtual machine?"),
                _("Virtual machine with the same name already exists, "
                  "it will be overwritten"))
            if not res:
                return
            self.overwrite = True
        else:
            res = self.err.yes_no(
                _("Are you sure you want to restore virtual machine?"),
                _("New virtual machine will be created"))
            if not res:
                return

        vmmAsyncJob.simple_async(self._async_restore,
                        [new_name], self, _("Restoring virtual machine"),
                        (_("This may take a while")),
                        (_("Error restoring virtual machine: '%s'")
                         % self.vm.get_name()), errorcb=self.errorcb)
        self.close()

    def errorcb(self, error, details):
        error = (_("Error during virtual machine restore '%(vm)s': "
                    "%(error)s") % {
                    "vm": self.vm.get_name(),
                    "error": error,
                    })
        self.err.show_err(error, details=details)

    def _init_new_guest(self, new_name, src_xml, conn):
        """
        Perform the series of unconditional new VM changes we always make
        """
        src_xml = src_xml.replace(self.vm.get_name(), new_name)
        self._new_guest = Guest(conn, parsexml=src_xml)
        self._new_guest.uuid = None
        self._new_guest.uuid = Guest.generate_uuid(conn)
        if not self.overwrite:
            self._new_guest.set_backup_schedule(None)
            self._new_guest.set_backup_repeat(None)
            self._new_guest.set_backup_folder(None)
            self._new_guest.set_backup_type(None)
            self._new_guest.set_backup_threshold(None)
            self._new_guest.set_backup_excluded(None)
            self._new_guest.set_backup_compress(None)
            self._new_guest.set_backup_included(None)

        for iface in self._new_guest.devices.interface:
            iface.target_dev = None
            iface.macaddr = DeviceInterface.generate_mac(conn)

    def save_xml(self, conn, new_name, file_name):
        with open(os.path.join(self.output, file_name), "r") as f:
            src_xml = f.read()
            self._init_new_guest(new_name, src_xml, conn)

    def save_disk(self, conn, pool, file_name, new_name):
        vm_name = self.vm.get_name()
        vol_name = file_name.replace(vm_name, new_name)
        if not self.overwrite:
            if vol_name == file_name:
                vol_name = file_name.replace(
                    ".qcow2", f"-{vm_name}-replace.qcow2")
                try:
                    vol = pool.storageVolLookupByName(vol_name)
                    raise Exception(_("Storage with name '%s' already exist")
                                    % vol_name)
                except:
                    pass

            for disk in self._new_guest.devices.disk:
                disk_name = disk.source.file.rsplit('/', 1)[1]
                if disk_name == file_name:
                    disk.source.file = disk.source.file.replace(disk_name,
                                                                vol_name)

        st = conn.newStream()
        cap = .000001

        vol = StorageVolume(conn)
        vol.pool = pool
        vol.name = vol_name
        vol.capacity = (cap * 1024 * 1024 * 1024)
        vol.allocation = vol.capacity
        vol.install(meter=None)
        vol = pool.storageVolLookupByName(vol_name)

        _upload_disk(vol, st, os.path.join(self.output, file_name))
        return vol_name
